SET DBG_FUNC=FALSE

SET DBG_CNTX_INFO=TRUE
SET DBG_CNTX_WARN=TRUE
SET DBG_CNTX_PROFILE=FALSE
SET DBG_CNTX_ERROR=TRUE
SET DBG_CNTX_ALL=TRUE

SET DBG_PRIM_INFO=FALSE
SET DBG_PRIM_WARN=FALSE
SET DBG_PRIM_PROFILE=FALSE
SET DBG_PRIM_ERROR=TRUE
SET DBG_PRIM_ALL=FALSE

SET DBG_TXT_INFO=FALSE
SET DBG_TXT_WARN=TRUE
SET DBG_TXT_PROFILE=FALSE
SET DBG_TXT_ERROR=TRUE
SET DBG_TXT_ALL=FALSE

SET DBG_ALL_INFO=FALSE
SET DBG_ALL_WARN=TRUE
SET DBG_ALL_PROFILE=FALSE
SET DBG_ALL_ERROR=TRUE
SET DBG_ALL=FALSE
