/* autogenerated by bdf2c!  do not edit */

/* "Copyright (c) 1984, 1987 Adobe Systems Incorporated. All Rights Reserved. Copyright (c) 1988, 1991 Digital Equipment Corporation. All Rights Reserved." */


#include "internal.h"
/*
typedef struct {
   int width, height;
   int xorig, yorig;
   int xmove;
   const unsigned char *bitmap;
} GLUTBitmapChar;

typedef struct {
   const char *name;
   int height;
   int num;
   const GLUTBitmapChar *const *table;
} GLUTBitmapFont;
*/


static const unsigned char ch32data[] = {
   0x0
};
static const GLUTBitmapChar ch32 = { 1, 1, 0, 0, 3, ch32data };

static const unsigned char ch33data[] = {
   0x80,0x0,0x80,0x80,0x80,0x80,0x80,0x80
};
static const GLUTBitmapChar ch33 = { 1, 8, -1, 0, 3, ch33data };

static const unsigned char ch34data[] = {
   0xa0,0xa0
};
static const GLUTBitmapChar ch34 = { 3, 2, -1, -6, 4, ch34data };

static const unsigned char ch35data[] = {
   0x50,0x50,0xf8,0x28,0x7c,0x28,0x28
};
static const GLUTBitmapChar ch35 = { 6, 7, 0, 0, 6, ch35data };

static const unsigned char ch36data[] = {
   0x20,0x70,0xa8,0x28,0x70,0xa0,0xa8,0x70,0x20
};
static const GLUTBitmapChar ch36 = { 5, 9, 0, 1, 6, ch36data };

static const unsigned char ch37data[] = {
   0x26,0x29,0x16,0x10,0x8,0x68,0x94,0x64
};
static const GLUTBitmapChar ch37 = { 8, 8, 0, 0, 9, ch37data };

static const unsigned char ch38data[] = {
   0x64,0x98,0x98,0xa4,0x60,0x50,0x50,0x20
};
static const GLUTBitmapChar ch38 = { 6, 8, -1, 0, 8, ch38data };

static const unsigned char ch39data[] = {
   0x80,0x40,0x40
};
static const GLUTBitmapChar ch39 = { 2, 3, -1, -5, 3, ch39data };

static const unsigned char ch40data[] = {
   0x20,0x40,0x40,0x80,0x80,0x80,0x80,0x40,0x40,0x20
};
static const GLUTBitmapChar ch40 = { 3, 10, 0, 2, 4, ch40data };

static const unsigned char ch41data[] = {
   0x80,0x40,0x40,0x20,0x20,0x20,0x20,0x40,0x40,0x80
};
static const GLUTBitmapChar ch41 = { 3, 10, -1, 2, 4, ch41data };

static const unsigned char ch42data[] = {
   0xa0,0x40,0xa0
};
static const GLUTBitmapChar ch42 = { 3, 3, 0, -5, 4, ch42data };

static const unsigned char ch43data[] = {
   0x20,0x20,0xf8,0x20,0x20
};
static const GLUTBitmapChar ch43 = { 5, 5, 0, -1, 6, ch43data };

static const unsigned char ch44data[] = {
   0x80,0x40,0x40
};
static const GLUTBitmapChar ch44 = { 2, 3, 0, 2, 3, ch44data };

static const unsigned char ch45data[] = {
   0xf8
};
static const GLUTBitmapChar ch45 = { 5, 1, -1, -3, 7, ch45data };

static const unsigned char ch46data[] = {
   0x80
};
static const GLUTBitmapChar ch46 = { 1, 1, -1, 0, 3, ch46data };

static const unsigned char ch47data[] = {
   0x80,0x80,0x40,0x40,0x40,0x40,0x20,0x20
};
static const GLUTBitmapChar ch47 = { 3, 8, 0, 0, 3, ch47data };

static const unsigned char ch48data[] = {
   0x70,0x88,0x88,0x88,0x88,0x88,0x88,0x70
};
static const GLUTBitmapChar ch48 = { 5, 8, 0, 0, 6, ch48data };

static const unsigned char ch49data[] = {
   0x40,0x40,0x40,0x40,0x40,0x40,0xc0,0x40
};
static const GLUTBitmapChar ch49 = { 2, 8, -1, 0, 6, ch49data };

static const unsigned char ch50data[] = {
   0xf8,0x80,0x40,0x30,0x8,0x8,0x88,0x70
};
static const GLUTBitmapChar ch50 = { 5, 8, 0, 0, 6, ch50data };

static const unsigned char ch51data[] = {
   0x70,0x88,0x8,0x8,0x30,0x8,0x88,0x70
};
static const GLUTBitmapChar ch51 = { 5, 8, 0, 0, 6, ch51data };

static const unsigned char ch52data[] = {
   0x10,0x10,0xf8,0x90,0x50,0x50,0x30,0x10
};
static const GLUTBitmapChar ch52 = { 5, 8, 0, 0, 6, ch52data };

static const unsigned char ch53data[] = {
   0x70,0x88,0x8,0x8,0xf0,0x80,0x80,0xf8
};
static const GLUTBitmapChar ch53 = { 5, 8, 0, 0, 6, ch53data };

static const unsigned char ch54data[] = {
   0x70,0x88,0x88,0xc8,0xb0,0x80,0x88,0x70
};
static const GLUTBitmapChar ch54 = { 5, 8, 0, 0, 6, ch54data };

static const unsigned char ch55data[] = {
   0x40,0x40,0x20,0x20,0x10,0x10,0x8,0xf8
};
static const GLUTBitmapChar ch55 = { 5, 8, 0, 0, 6, ch55data };

static const unsigned char ch56data[] = {
   0x70,0x88,0x88,0x88,0x70,0x88,0x88,0x70
};
static const GLUTBitmapChar ch56 = { 5, 8, 0, 0, 6, ch56data };

static const unsigned char ch57data[] = {
   0x70,0x88,0x8,0x68,0x98,0x88,0x88,0x70
};
static const GLUTBitmapChar ch57 = { 5, 8, 0, 0, 6, ch57data };

static const unsigned char ch58data[] = {
   0x80,0x0,0x0,0x0,0x0,0x80
};
static const GLUTBitmapChar ch58 = { 1, 6, -1, 0, 3, ch58data };

static const unsigned char ch59data[] = {
   0x80,0x40,0x40,0x0,0x0,0x0,0x0,0x40
};
static const GLUTBitmapChar ch59 = { 2, 8, 0, 2, 3, ch59data };

static const unsigned char ch60data[] = {
   0x20,0x40,0x80,0x40,0x20
};
static const GLUTBitmapChar ch60 = { 3, 5, -1, -1, 6, ch60data };

static const unsigned char ch61data[] = {
   0xf0,0x0,0xf0
};
static const GLUTBitmapChar ch61 = { 4, 3, 0, -2, 5, ch61data };

static const unsigned char ch62data[] = {
   0x80,0x40,0x20,0x40,0x80
};
static const GLUTBitmapChar ch62 = { 3, 5, -1, -1, 6, ch62data };

static const unsigned char ch63data[] = {
   0x40,0x0,0x40,0x40,0x20,0x10,0x90,0x60
};
static const GLUTBitmapChar ch63 = { 4, 8, -1, 0, 6, ch63data };

static const unsigned char ch64data[] = {
   0x3e,0x0,0x40,0x0,0x9b,0x0,0xa4,0x80,0xa4,0x80,0xa2,0x40,0x92,0x40,0x4d,0x40,
   0x20,0x80,0x1f,0x0
};
static const GLUTBitmapChar ch64 = { 10, 10, 0, 2, 11, ch64data };

static const unsigned char ch65data[] = {
   0x82,0x82,0x7c,0x44,0x28,0x28,0x10,0x10
};
static const GLUTBitmapChar ch65 = { 7, 8, 0, 0, 7, ch65data };

static const unsigned char ch66data[] = {
   0xf0,0x88,0x88,0x88,0xf0,0x88,0x88,0xf0
};
static const GLUTBitmapChar ch66 = { 5, 8, -1, 0, 7, ch66data };

static const unsigned char ch67data[] = {
   0x78,0x84,0x80,0x80,0x80,0x80,0x84,0x78
};
static const GLUTBitmapChar ch67 = { 6, 8, -1, 0, 8, ch67data };

static const unsigned char ch68data[] = {
   0xf0,0x88,0x84,0x84,0x84,0x84,0x88,0xf0
};
static const GLUTBitmapChar ch68 = { 6, 8, -1, 0, 8, ch68data };

static const unsigned char ch69data[] = {
   0xf8,0x80,0x80,0x80,0xf8,0x80,0x80,0xf8
};
static const GLUTBitmapChar ch69 = { 5, 8, -1, 0, 7, ch69data };

static const unsigned char ch70data[] = {
   0x80,0x80,0x80,0x80,0xf0,0x80,0x80,0xf8
};
static const GLUTBitmapChar ch70 = { 5, 8, -1, 0, 6, ch70data };

static const unsigned char ch71data[] = {
   0x74,0x8c,0x84,0x8c,0x80,0x80,0x84,0x78
};
static const GLUTBitmapChar ch71 = { 6, 8, -1, 0, 8, ch71data };

static const unsigned char ch72data[] = {
   0x84,0x84,0x84,0x84,0xfc,0x84,0x84,0x84
};
static const GLUTBitmapChar ch72 = { 6, 8, -1, 0, 8, ch72data };

static const unsigned char ch73data[] = {
   0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x80
};
static const GLUTBitmapChar ch73 = { 1, 8, -1, 0, 3, ch73data };

static const unsigned char ch74data[] = {
   0x60,0x90,0x10,0x10,0x10,0x10,0x10,0x10
};
static const GLUTBitmapChar ch74 = { 4, 8, 0, 0, 5, ch74data };

static const unsigned char ch75data[] = {
   0x88,0x88,0x90,0x90,0xe0,0xa0,0x90,0x88
};
static const GLUTBitmapChar ch75 = { 5, 8, -1, 0, 7, ch75data };

static const unsigned char ch76data[] = {
   0xf0,0x80,0x80,0x80,0x80,0x80,0x80,0x80
};
static const GLUTBitmapChar ch76 = { 4, 8, -1, 0, 6, ch76data };

static const unsigned char ch77data[] = {
   0x92,0x92,0x92,0xaa,0xaa,0xc6,0xc6,0x82
};
static const GLUTBitmapChar ch77 = { 7, 8, -1, 0, 9, ch77data };

static const unsigned char ch78data[] = {
   0x8c,0x8c,0x94,0x94,0xa4,0xa4,0xc4,0xc4
};
static const GLUTBitmapChar ch78 = { 6, 8, -1, 0, 8, ch78data };

static const unsigned char ch79data[] = {
   0x78,0x84,0x84,0x84,0x84,0x84,0x84,0x78
};
static const GLUTBitmapChar ch79 = { 6, 8, -1, 0, 8, ch79data };

static const unsigned char ch80data[] = {
   0x80,0x80,0x80,0x80,0xf0,0x88,0x88,0xf0
};
static const GLUTBitmapChar ch80 = { 5, 8, -1, 0, 7, ch80data };

static const unsigned char ch81data[] = {
   0x2,0x7c,0x8c,0x94,0x84,0x84,0x84,0x84,0x78
};
static const GLUTBitmapChar ch81 = { 7, 9, -1, 1, 8, ch81data };

static const unsigned char ch82data[] = {
   0x88,0x88,0x88,0x88,0xf0,0x88,0x88,0xf0
};
static const GLUTBitmapChar ch82 = { 5, 8, -1, 0, 7, ch82data };

static const unsigned char ch83data[] = {
   0x70,0x88,0x88,0x8,0x70,0x80,0x88,0x70
};
static const GLUTBitmapChar ch83 = { 5, 8, -1, 0, 7, ch83data };

static const unsigned char ch84data[] = {
   0x20,0x20,0x20,0x20,0x20,0x20,0x20,0xf8
};
static const GLUTBitmapChar ch84 = { 5, 8, 0, 0, 5, ch84data };

static const unsigned char ch85data[] = {
   0x78,0x84,0x84,0x84,0x84,0x84,0x84,0x84
};
static const GLUTBitmapChar ch85 = { 6, 8, -1, 0, 8, ch85data };

static const unsigned char ch86data[] = {
   0x10,0x28,0x28,0x44,0x44,0x44,0x82,0x82
};
static const GLUTBitmapChar ch86 = { 7, 8, 0, 0, 7, ch86data };

static const unsigned char ch87data[] = {
   0x22,0x0,0x22,0x0,0x22,0x0,0x55,0x0,0x49,0x0,0x49,0x0,0x88,0x80,0x88,0x80
};
static const GLUTBitmapChar ch87 = { 9, 8, 0, 0, 9, ch87data };

static const unsigned char ch88data[] = {
   0x88,0x88,0x50,0x50,0x20,0x50,0x88,0x88
};
static const GLUTBitmapChar ch88 = { 5, 8, -1, 0, 7, ch88data };

static const unsigned char ch89data[] = {
   0x10,0x10,0x10,0x28,0x28,0x44,0x44,0x82
};
static const GLUTBitmapChar ch89 = { 7, 8, 0, 0, 7, ch89data };

static const unsigned char ch90data[] = {
   0xf8,0x80,0x40,0x20,0x20,0x10,0x8,0xf8
};
static const GLUTBitmapChar ch90 = { 5, 8, -1, 0, 7, ch90data };

static const unsigned char ch91data[] = {
   0xc0,0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x80,0xc0
};
static const GLUTBitmapChar ch91 = { 2, 10, -1, 2, 3, ch91data };

static const unsigned char ch92data[] = {
   0x20,0x20,0x40,0x40,0x40,0x40,0x80,0x80
};
static const GLUTBitmapChar ch92 = { 3, 8, 0, 0, 3, ch92data };

static const unsigned char ch93data[] = {
   0xc0,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0xc0
};
static const GLUTBitmapChar ch93 = { 2, 10, 0, 2, 3, ch93data };

static const unsigned char ch94data[] = {
   0x88,0x50,0x50,0x20,0x20
};
static const GLUTBitmapChar ch94 = { 5, 5, 0, -3, 6, ch94data };

static const unsigned char ch95data[] = {
   0xfc
};
static const GLUTBitmapChar ch95 = { 6, 1, 0, 2, 6, ch95data };

static const unsigned char ch96data[] = {
   0x80,0x80,0x40
};
static const GLUTBitmapChar ch96 = { 2, 3, 0, -5, 3, ch96data };

static const unsigned char ch97data[] = {
   0x68,0x90,0x90,0x70,0x10,0xe0
};
static const GLUTBitmapChar ch97 = { 5, 6, 0, 0, 5, ch97data };

static const unsigned char ch98data[] = {
   0xb0,0xc8,0x88,0x88,0xc8,0xb0,0x80,0x80
};
static const GLUTBitmapChar ch98 = { 5, 8, 0, 0, 6, ch98data };

static const unsigned char ch99data[] = {
   0x60,0x90,0x80,0x80,0x90,0x60
};
static const GLUTBitmapChar ch99 = { 4, 6, 0, 0, 5, ch99data };

static const unsigned char ch100data[] = {
   0x68,0x98,0x88,0x88,0x98,0x68,0x8,0x8
};
static const GLUTBitmapChar ch100 = { 5, 8, 0, 0, 6, ch100data };

static const unsigned char ch101data[] = {
   0x60,0x90,0x80,0xf0,0x90,0x60
};
static const GLUTBitmapChar ch101 = { 4, 6, 0, 0, 5, ch101data };

static const unsigned char ch102data[] = {
   0x40,0x40,0x40,0x40,0x40,0xe0,0x40,0x30
};
static const GLUTBitmapChar ch102 = { 4, 8, 0, 0, 4, ch102data };

static const unsigned char ch103data[] = {
   0x70,0x8,0x68,0x98,0x88,0x88,0x98,0x68
};
static const GLUTBitmapChar ch103 = { 5, 8, 0, 2, 6, ch103data };

static const unsigned char ch104data[] = {
   0x88,0x88,0x88,0x88,0xc8,0xb0,0x80,0x80
};
static const GLUTBitmapChar ch104 = { 5, 8, 0, 0, 6, ch104data };

static const unsigned char ch105data[] = {
   0x80,0x80,0x80,0x80,0x80,0x80,0x0,0x80
};
static const GLUTBitmapChar ch105 = { 1, 8, 0, 0, 2, ch105data };

static const unsigned char ch106data[] = {
   0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x0,0x80
};
static const GLUTBitmapChar ch106 = { 1, 9, 0, 1, 2, ch106data };

static const unsigned char ch107data[] = {
   0x90,0x90,0xa0,0xc0,0xa0,0x90,0x80,0x80
};
static const GLUTBitmapChar ch107 = { 4, 8, 0, 0, 5, ch107data };

static const unsigned char ch108data[] = {
   0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x80
};
static const GLUTBitmapChar ch108 = { 1, 8, 0, 0, 2, ch108data };

static const unsigned char ch109data[] = {
   0x92,0x92,0x92,0x92,0x92,0xec
};
static const GLUTBitmapChar ch109 = { 7, 6, 0, 0, 8, ch109data };

static const unsigned char ch110data[] = {
   0x88,0x88,0x88,0x88,0xc8,0xb0
};
static const GLUTBitmapChar ch110 = { 5, 6, 0, 0, 6, ch110data };

static const unsigned char ch111data[] = {
   0x70,0x88,0x88,0x88,0x88,0x70
};
static const GLUTBitmapChar ch111 = { 5, 6, 0, 0, 6, ch111data };

static const unsigned char ch112data[] = {
   0x80,0x80,0xb0,0xc8,0x88,0x88,0xc8,0xb0
};
static const GLUTBitmapChar ch112 = { 5, 8, 0, 2, 6, ch112data };

static const unsigned char ch113data[] = {
   0x8,0x8,0x68,0x98,0x88,0x88,0x98,0x68
};
static const GLUTBitmapChar ch113 = { 5, 8, 0, 2, 6, ch113data };

static const unsigned char ch114data[] = {
   0x80,0x80,0x80,0x80,0xc0,0xa0
};
static const GLUTBitmapChar ch114 = { 3, 6, 0, 0, 4, ch114data };

static const unsigned char ch115data[] = {
   0x60,0x90,0x10,0x60,0x90,0x60
};
static const GLUTBitmapChar ch115 = { 4, 6, 0, 0, 5, ch115data };

static const unsigned char ch116data[] = {
   0x60,0x40,0x40,0x40,0x40,0xe0,0x40,0x40
};
static const GLUTBitmapChar ch116 = { 3, 8, 0, 0, 4, ch116data };

static const unsigned char ch117data[] = {
   0x70,0x90,0x90,0x90,0x90,0x90
};
static const GLUTBitmapChar ch117 = { 4, 6, 0, 0, 5, ch117data };

static const unsigned char ch118data[] = {
   0x20,0x20,0x50,0x50,0x88,0x88
};
static const GLUTBitmapChar ch118 = { 5, 6, 0, 0, 6, ch118data };

static const unsigned char ch119data[] = {
   0x28,0x28,0x54,0x54,0x92,0x92
};
static const GLUTBitmapChar ch119 = { 7, 6, 0, 0, 8, ch119data };

static const unsigned char ch120data[] = {
   0x88,0x88,0x50,0x20,0x50,0x88
};
static const GLUTBitmapChar ch120 = { 5, 6, 0, 0, 6, ch120data };

static const unsigned char ch121data[] = {
   0x80,0x40,0x40,0x60,0xa0,0xa0,0x90,0x90
};
static const GLUTBitmapChar ch121 = { 4, 8, 0, 2, 5, ch121data };

static const unsigned char ch122data[] = {
   0xf0,0x80,0x40,0x20,0x10,0xf0
};
static const GLUTBitmapChar ch122 = { 4, 6, 0, 0, 5, ch122data };

static const unsigned char ch123data[] = {
   0x20,0x40,0x40,0x40,0x40,0x80,0x40,0x40,0x40,0x20
};
static const GLUTBitmapChar ch123 = { 3, 10, 0, 2, 3, ch123data };

static const unsigned char ch124data[] = {
   0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x80
};
static const GLUTBitmapChar ch124 = { 1, 10, -1, 2, 3, ch124data };

static const unsigned char ch125data[] = {
   0x80,0x40,0x40,0x40,0x40,0x20,0x40,0x40,0x40,0x80
};
static const GLUTBitmapChar ch125 = { 3, 10, 0, 2, 3, ch125data };

static const unsigned char ch126data[] = {
   0x98,0x64
};
static const GLUTBitmapChar ch126 = { 6, 2, 0, -3, 7, ch126data };

static const unsigned char ch160data[] = {
   0x0
};
static const GLUTBitmapChar ch160 = { 1, 1, 0, 0, 3, ch160data };

static const unsigned char ch161data[] = {
   0x80,0x80,0x80,0x80,0x80,0x80,0x0,0x80
};
static const GLUTBitmapChar ch161 = { 1, 8, -1, 2, 3, ch161data };

static const unsigned char ch162data[] = {
   0x40,0x70,0xa8,0xa0,0xa0,0xa8,0x70,0x10
};
static const GLUTBitmapChar ch162 = { 5, 8, 0, 1, 6, ch162data };

static const unsigned char ch163data[] = {
   0xb0,0x48,0x40,0x40,0xe0,0x40,0x48,0x30
};
static const GLUTBitmapChar ch163 = { 5, 8, 0, 0, 6, ch163data };

static const unsigned char ch164data[] = {
   0x90,0x60,0x90,0x90,0x60,0x90
};
static const GLUTBitmapChar ch164 = { 4, 6, 0, -1, 5, ch164data };

static const unsigned char ch165data[] = {
   0x20,0xf8,0x20,0xf8,0x50,0x50,0x88,0x88
};
static const GLUTBitmapChar ch165 = { 5, 8, 0, 0, 6, ch165data };

static const unsigned char ch166data[] = {
   0x80,0x80,0x80,0x80,0x0,0x0,0x80,0x80,0x80,0x80
};
static const GLUTBitmapChar ch166 = { 1, 10, -1, 2, 3, ch166data };

static const unsigned char ch167data[] = {
   0x70,0x88,0x18,0x70,0xc8,0x98,0x70,0xc0,0x88,0x70
};
static const GLUTBitmapChar ch167 = { 5, 10, 0, 2, 6, ch167data };

static const unsigned char ch168data[] = {
   0xa0
};
static const GLUTBitmapChar ch168 = { 3, 1, 0, -7, 3, ch168data };

static const unsigned char ch169data[] = {
   0x38,0x44,0x9a,0xa2,0x9a,0x44,0x38
};
static const GLUTBitmapChar ch169 = { 7, 7, -1, 0, 9, ch169data };

static const unsigned char ch170data[] = {
   0xe0,0x0,0xa0,0x20,0xe0
};
static const GLUTBitmapChar ch170 = { 3, 5, 0, -3, 4, ch170data };

static const unsigned char ch171data[] = {
   0x28,0x50,0xa0,0x50,0x28
};
static const GLUTBitmapChar ch171 = { 5, 5, 0, 0, 6, ch171data };

static const unsigned char ch172data[] = {
   0x8,0x8,0xf8
};
static const GLUTBitmapChar ch172 = { 5, 3, -1, -2, 7, ch172data };

static const unsigned char ch173data[] = {
   0xe0
};
static const GLUTBitmapChar ch173 = { 3, 1, 0, -3, 4, ch173data };

static const unsigned char ch174data[] = {
   0x38,0x44,0xaa,0xb2,0xba,0x44,0x38
};
static const GLUTBitmapChar ch174 = { 7, 7, -1, 0, 9, ch174data };

static const unsigned char ch175data[] = {
   0xe0
};
static const GLUTBitmapChar ch175 = { 3, 1, 0, -7, 3, ch175data };

static const unsigned char ch176data[] = {
   0x60,0x90,0x90,0x60
};
static const GLUTBitmapChar ch176 = { 4, 4, 0, -3, 4, ch176data };

static const unsigned char ch177data[] = {
   0xf8,0x0,0x20,0x20,0xf8,0x20,0x20
};
static const GLUTBitmapChar ch177 = { 5, 7, 0, 0, 6, ch177data };

static const unsigned char ch178data[] = {
   0xe0,0x40,0xa0,0x60
};
static const GLUTBitmapChar ch178 = { 3, 4, 0, -3, 3, ch178data };

static const unsigned char ch179data[] = {
   0xc0,0x20,0x40,0xe0
};
static const GLUTBitmapChar ch179 = { 3, 4, 0, -3, 3, ch179data };

static const unsigned char ch180data[] = {
   0x80,0x40
};
static const GLUTBitmapChar ch180 = { 2, 2, 0, -6, 3, ch180data };

static const unsigned char ch181data[] = {
   0x80,0x80,0xf0,0x90,0x90,0x90,0x90,0x90
};
static const GLUTBitmapChar ch181 = { 4, 8, 0, 2, 5, ch181data };

static const unsigned char ch182data[] = {
   0x28,0x28,0x28,0x28,0x28,0x68,0xe8,0xe8,0xe8,0x7c
};
static const GLUTBitmapChar ch182 = { 6, 10, 0, 2, 6, ch182data };

static const unsigned char ch183data[] = {
   0xc0
};
static const GLUTBitmapChar ch183 = { 2, 1, 0, -3, 3, ch183data };

static const unsigned char ch184data[] = {
   0xc0,0x40
};
static const GLUTBitmapChar ch184 = { 2, 2, 0, 2, 3, ch184data };

static const unsigned char ch185data[] = {
   0x40,0x40,0xc0,0x40
};
static const GLUTBitmapChar ch185 = { 2, 4, 0, -3, 3, ch185data };

static const unsigned char ch186data[] = {
   0xe0,0x0,0xe0,0xa0,0xe0
};
static const GLUTBitmapChar ch186 = { 3, 5, 0, -3, 4, ch186data };

static const unsigned char ch187data[] = {
   0xa0,0x50,0x28,0x50,0xa0
};
static const GLUTBitmapChar ch187 = { 5, 5, 0, 0, 6, ch187data };

static const unsigned char ch188data[] = {
   0x21,0x0,0x17,0x80,0x13,0x0,0x9,0x0,0x48,0x0,0x44,0x0,0xc4,0x0,0x42,0x0
};
static const GLUTBitmapChar ch188 = { 9, 8, 0, 0, 9, ch188data };

static const unsigned char ch189data[] = {
   0x27,0x12,0x15,0xb,0x48,0x44,0xc4,0x42
};
static const GLUTBitmapChar ch189 = { 8, 8, 0, 0, 9, ch189data };

static const unsigned char ch190data[] = {
   0x21,0x0,0x17,0x80,0x13,0x0,0x9,0x0,0xc8,0x0,0x24,0x0,0x44,0x0,0xe2,0x0
};
static const GLUTBitmapChar ch190 = { 9, 8, 0, 0, 9, ch190data };

static const unsigned char ch191data[] = {
   0x60,0x90,0x80,0x40,0x20,0x20,0x0,0x20
};
static const GLUTBitmapChar ch191 = { 4, 8, -1, 2, 6, ch191data };

static const unsigned char ch192data[] = {
   0x82,0x82,0x7c,0x44,0x28,0x28,0x10,0x10,0x0,0x10,0x20
};
static const GLUTBitmapChar ch192 = { 7, 11, 0, 0, 7, ch192data };

static const unsigned char ch193data[] = {
   0x82,0x82,0x7c,0x44,0x28,0x28,0x10,0x10,0x0,0x10,0x8
};
static const GLUTBitmapChar ch193 = { 7, 11, 0, 0, 7, ch193data };

static const unsigned char ch194data[] = {
   0x82,0x82,0x7c,0x44,0x28,0x28,0x10,0x10,0x0,0x28,0x10
};
static const GLUTBitmapChar ch194 = { 7, 11, 0, 0, 7, ch194data };

static const unsigned char ch195data[] = {
   0x82,0x82,0x7c,0x44,0x28,0x28,0x10,0x10,0x0,0x28,0x14
};
static const GLUTBitmapChar ch195 = { 7, 11, 0, 0, 7, ch195data };

static const unsigned char ch196data[] = {
   0x82,0x82,0x7c,0x44,0x28,0x28,0x10,0x10,0x0,0x28
};
static const GLUTBitmapChar ch196 = { 7, 10, 0, 0, 7, ch196data };

static const unsigned char ch197data[] = {
   0x82,0x82,0x7c,0x44,0x28,0x28,0x10,0x10,0x10,0x28,0x10
};
static const GLUTBitmapChar ch197 = { 7, 11, 0, 0, 7, ch197data };

static const unsigned char ch198data[] = {
   0x8f,0x80,0x88,0x0,0x78,0x0,0x48,0x0,0x2f,0x80,0x28,0x0,0x18,0x0,0x1f,0x80
};
static const GLUTBitmapChar ch198 = { 9, 8, 0, 0, 10, ch198data };

static const unsigned char ch199data[] = {
   0x30,0x10,0x78,0x84,0x80,0x80,0x80,0x80,0x84,0x78
};
static const GLUTBitmapChar ch199 = { 6, 10, -1, 2, 8, ch199data };

static const unsigned char ch200data[] = {
   0xf8,0x80,0x80,0x80,0xf8,0x80,0x80,0xf8,0x0,0x20,0x40
};
static const GLUTBitmapChar ch200 = { 5, 11, -1, 0, 7, ch200data };

static const unsigned char ch201data[] = {
   0xf8,0x80,0x80,0x80,0xf8,0x80,0x80,0xf8,0x0,0x20,0x10
};
static const GLUTBitmapChar ch201 = { 5, 11, -1, 0, 7, ch201data };

static const unsigned char ch202data[] = {
   0xf8,0x80,0x80,0xf8,0x80,0x80,0x80,0xf8,0x0,0x50,0x20
};
static const GLUTBitmapChar ch202 = { 5, 11, -1, 0, 7, ch202data };

static const unsigned char ch203data[] = {
   0xf8,0x80,0x80,0x80,0xf8,0x80,0x80,0xf8,0x0,0x50
};
static const GLUTBitmapChar ch203 = { 5, 10, -1, 0, 7, ch203data };

static const unsigned char ch204data[] = {
   0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x0,0x40,0x80
};
static const GLUTBitmapChar ch204 = { 2, 11, 0, 0, 3, ch204data };

static const unsigned char ch205data[] = {
   0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x0,0x80,0x40
};
static const GLUTBitmapChar ch205 = { 2, 11, -1, 0, 3, ch205data };

static const unsigned char ch206data[] = {
   0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x0,0xa0,0x40
};
static const GLUTBitmapChar ch206 = { 3, 11, 0, 0, 3, ch206data };

static const unsigned char ch207data[] = {
   0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x40,0x0,0xa0
};
static const GLUTBitmapChar ch207 = { 3, 10, 0, 0, 3, ch207data };

static const unsigned char ch208data[] = {
   0x78,0x44,0x42,0x42,0xf2,0x42,0x44,0x78
};
static const GLUTBitmapChar ch208 = { 7, 8, 0, 0, 8, ch208data };

static const unsigned char ch209data[] = {
   0x8c,0x8c,0x94,0x94,0xa4,0xa4,0xc4,0xc4,0x0,0x50,0x28
};
static const GLUTBitmapChar ch209 = { 6, 11, -1, 0, 8, ch209data };

static const unsigned char ch210data[] = {
   0x78,0x84,0x84,0x84,0x84,0x84,0x84,0x78,0x0,0x10,0x20
};
static const GLUTBitmapChar ch210 = { 6, 11, -1, 0, 8, ch210data };

static const unsigned char ch211data[] = {
   0x78,0x84,0x84,0x84,0x84,0x84,0x84,0x78,0x0,0x10,0x8
};
static const GLUTBitmapChar ch211 = { 6, 11, -1, 0, 8, ch211data };

static const unsigned char ch212data[] = {
   0x78,0x84,0x84,0x84,0x84,0x84,0x84,0x78,0x0,0x28,0x10
};
static const GLUTBitmapChar ch212 = { 6, 11, -1, 0, 8, ch212data };

static const unsigned char ch213data[] = {
   0x78,0x84,0x84,0x84,0x84,0x84,0x84,0x78,0x0,0x50,0x28
};
static const GLUTBitmapChar ch213 = { 6, 11, -1, 0, 8, ch213data };

static const unsigned char ch214data[] = {
   0x78,0x84,0x84,0x84,0x84,0x84,0x84,0x78,0x0,0x48
};
static const GLUTBitmapChar ch214 = { 6, 10, -1, 0, 8, ch214data };

static const unsigned char ch215data[] = {
   0x88,0x50,0x20,0x50,0x88
};
static const GLUTBitmapChar ch215 = { 5, 5, 0, -1, 6, ch215data };

static const unsigned char ch216data[] = {
   0x80,0x78,0xc4,0xa4,0xa4,0x94,0x94,0x8c,0x78,0x4
};
static const GLUTBitmapChar ch216 = { 6, 10, -1, 1, 8, ch216data };

static const unsigned char ch217data[] = {
   0x78,0x84,0x84,0x84,0x84,0x84,0x84,0x84,0x0,0x10,0x20
};
static const GLUTBitmapChar ch217 = { 6, 11, -1, 0, 8, ch217data };

static const unsigned char ch218data[] = {
   0x78,0x84,0x84,0x84,0x84,0x84,0x84,0x84,0x0,0x20,0x10
};
static const GLUTBitmapChar ch218 = { 6, 11, -1, 0, 8, ch218data };

static const unsigned char ch219data[] = {
   0x78,0x84,0x84,0x84,0x84,0x84,0x84,0x84,0x0,0x28,0x10
};
static const GLUTBitmapChar ch219 = { 6, 11, -1, 0, 8, ch219data };

static const unsigned char ch220data[] = {
   0x78,0x84,0x84,0x84,0x84,0x84,0x84,0x84,0x0,0x48
};
static const GLUTBitmapChar ch220 = { 6, 10, -1, 0, 8, ch220data };

static const unsigned char ch221data[] = {
   0x10,0x10,0x10,0x28,0x28,0x44,0x44,0x82,0x0,0x10,0x8
};
static const GLUTBitmapChar ch221 = { 7, 11, 0, 0, 7, ch221data };

static const unsigned char ch222data[] = {
   0x80,0x80,0xf0,0x88,0x88,0xf0,0x80,0x80
};
static const GLUTBitmapChar ch222 = { 5, 8, -1, 0, 7, ch222data };

static const unsigned char ch223data[] = {
   0xa0,0x90,0x90,0x90,0xa0,0x90,0x90,0x60
};
static const GLUTBitmapChar ch223 = { 4, 8, 0, 0, 5, ch223data };

static const unsigned char ch224data[] = {
   0x68,0x90,0x90,0x70,0x10,0xe0,0x0,0x20,0x40
};
static const GLUTBitmapChar ch224 = { 5, 9, 0, 0, 5, ch224data };

static const unsigned char ch225data[] = {
   0x68,0x90,0x90,0x70,0x10,0xe0,0x0,0x20,0x10
};
static const GLUTBitmapChar ch225 = { 5, 9, 0, 0, 5, ch225data };

static const unsigned char ch226data[] = {
   0x68,0x90,0x90,0x70,0x10,0xe0,0x0,0x50,0x20
};
static const GLUTBitmapChar ch226 = { 5, 9, 0, 0, 5, ch226data };

static const unsigned char ch227data[] = {
   0x68,0x90,0x90,0x70,0x10,0xe0,0x0,0xa0,0x50
};
static const GLUTBitmapChar ch227 = { 5, 9, 0, 0, 5, ch227data };

static const unsigned char ch228data[] = {
   0x68,0x90,0x90,0x70,0x10,0xe0,0x0,0x50
};
static const GLUTBitmapChar ch228 = { 5, 8, 0, 0, 5, ch228data };

static const unsigned char ch229data[] = {
   0x68,0x90,0x90,0x70,0x10,0xe0,0x20,0x50,0x20
};
static const GLUTBitmapChar ch229 = { 5, 9, 0, 0, 5, ch229data };

static const unsigned char ch230data[] = {
   0x6c,0x92,0x90,0x7e,0x12,0xec
};
static const GLUTBitmapChar ch230 = { 7, 6, 0, 0, 8, ch230data };

static const unsigned char ch231data[] = {
   0x60,0x20,0x60,0x90,0x80,0x80,0x90,0x60
};
static const GLUTBitmapChar ch231 = { 4, 8, 0, 2, 5, ch231data };

static const unsigned char ch232data[] = {
   0x60,0x90,0x80,0xf0,0x90,0x60,0x0,0x20,0x40
};
static const GLUTBitmapChar ch232 = { 4, 9, 0, 0, 5, ch232data };

static const unsigned char ch233data[] = {
   0x60,0x90,0x80,0xf0,0x90,0x60,0x0,0x40,0x20
};
static const GLUTBitmapChar ch233 = { 4, 9, 0, 0, 5, ch233data };

static const unsigned char ch234data[] = {
   0x60,0x90,0x80,0xf0,0x90,0x60,0x0,0x50,0x20
};
static const GLUTBitmapChar ch234 = { 4, 9, 0, 0, 5, ch234data };

static const unsigned char ch235data[] = {
   0x60,0x90,0x80,0xf0,0x90,0x60,0x0,0x50
};
static const GLUTBitmapChar ch235 = { 4, 8, 0, 0, 5, ch235data };

static const unsigned char ch236data[] = {
   0x40,0x40,0x40,0x40,0x40,0x40,0x0,0x40,0x80
};
static const GLUTBitmapChar ch236 = { 2, 9, 1, 0, 2, ch236data };

static const unsigned char ch237data[] = {
   0x80,0x80,0x80,0x80,0x80,0x80,0x0,0x80,0x40
};
static const GLUTBitmapChar ch237 = { 2, 9, 0, 0, 2, ch237data };

static const unsigned char ch238data[] = {
   0x40,0x40,0x40,0x40,0x40,0x40,0x0,0xa0,0x40
};
static const GLUTBitmapChar ch238 = { 3, 9, 1, 0, 2, ch238data };

static const unsigned char ch239data[] = {
   0x40,0x40,0x40,0x40,0x40,0x40,0x0,0xa0
};
static const GLUTBitmapChar ch239 = { 3, 8, 0, 0, 2, ch239data };

static const unsigned char ch240data[] = {
   0x70,0x88,0x88,0x88,0x88,0x78,0x90,0x60,0x50
};
static const GLUTBitmapChar ch240 = { 5, 9, 0, 0, 6, ch240data };

static const unsigned char ch241data[] = {
   0x90,0x90,0x90,0x90,0x90,0xe0,0x0,0xa0,0x50
};
static const GLUTBitmapChar ch241 = { 4, 9, 0, 0, 5, ch241data };

static const unsigned char ch242data[] = {
   0x70,0x88,0x88,0x88,0x88,0x70,0x0,0x20,0x40
};
static const GLUTBitmapChar ch242 = { 5, 9, 0, 0, 6, ch242data };

static const unsigned char ch243data[] = {
   0x70,0x88,0x88,0x88,0x88,0x70,0x0,0x20,0x10
};
static const GLUTBitmapChar ch243 = { 5, 9, 0, 0, 6, ch243data };

static const unsigned char ch244data[] = {
   0x70,0x88,0x88,0x88,0x88,0x70,0x0,0x50,0x20
};
static const GLUTBitmapChar ch244 = { 5, 9, 0, 0, 6, ch244data };

static const unsigned char ch245data[] = {
   0x70,0x88,0x88,0x88,0x88,0x70,0x0,0x50,0x28
};
static const GLUTBitmapChar ch245 = { 5, 9, 0, 0, 6, ch245data };

static const unsigned char ch246data[] = {
   0x70,0x88,0x88,0x88,0x88,0x70,0x0,0x50
};
static const GLUTBitmapChar ch246 = { 5, 8, 0, 0, 6, ch246data };

static const unsigned char ch247data[] = {
   0x20,0x0,0xf8,0x0,0x20
};
static const GLUTBitmapChar ch247 = { 5, 5, 0, -1, 6, ch247data };

static const unsigned char ch248data[] = {
   0x70,0x88,0xc8,0xa8,0x98,0x74
};
static const GLUTBitmapChar ch248 = { 6, 6, 0, 0, 6, ch248data };

static const unsigned char ch249data[] = {
   0x70,0x90,0x90,0x90,0x90,0x90,0x0,0x20,0x40
};
static const GLUTBitmapChar ch249 = { 4, 9, 0, 0, 5, ch249data };

static const unsigned char ch250data[] = {
   0x70,0x90,0x90,0x90,0x90,0x90,0x0,0x40,0x20
};
static const GLUTBitmapChar ch250 = { 4, 9, 0, 0, 5, ch250data };

static const unsigned char ch251data[] = {
   0x70,0x90,0x90,0x90,0x90,0x90,0x0,0x50,0x20
};
static const GLUTBitmapChar ch251 = { 4, 9, 0, 0, 5, ch251data };

static const unsigned char ch252data[] = {
   0x70,0x90,0x90,0x90,0x90,0x90,0x0,0x50
};
static const GLUTBitmapChar ch252 = { 4, 8, 0, 0, 5, ch252data };

static const unsigned char ch253data[] = {
   0x80,0x40,0x40,0x60,0xa0,0xa0,0x90,0x90,0x0,0x20,0x10
};
static const GLUTBitmapChar ch253 = { 4, 11, 0, 2, 5, ch253data };

static const unsigned char ch254data[] = {
   0x80,0x80,0xb0,0xc8,0x88,0x88,0xc8,0xb0,0x80,0x80
};
static const GLUTBitmapChar ch254 = { 5, 10, 0, 2, 6, ch254data };

static const unsigned char ch255data[] = {
   0x80,0x40,0x40,0x60,0xa0,0xa0,0x90,0x90,0x0,0x50
};
static const GLUTBitmapChar ch255 = { 4, 10, 0, 2, 5, ch255data };


static const GLUTBitmapChar *chars[] = {
   0, 0, 0, 0, 0, 0, 0, 0, 
   0, 0, 0, 0, 0, 0, 0, 0, 
   0, 0, 0, 0, 0, 0, 0, 0, 
   0, 0, 0, 0, 0, 0, 0, 0, 
   &ch32, &ch33, &ch34, &ch35, &ch36, &ch37, &ch38, &ch39, 
   &ch40, &ch41, &ch42, &ch43, &ch44, &ch45, &ch46, &ch47, 
   &ch48, &ch49, &ch50, &ch51, &ch52, &ch53, &ch54, &ch55, 
   &ch56, &ch57, &ch58, &ch59, &ch60, &ch61, &ch62, &ch63, 
   &ch64, &ch65, &ch66, &ch67, &ch68, &ch69, &ch70, &ch71, 
   &ch72, &ch73, &ch74, &ch75, &ch76, &ch77, &ch78, &ch79, 
   &ch80, &ch81, &ch82, &ch83, &ch84, &ch85, &ch86, &ch87, 
   &ch88, &ch89, &ch90, &ch91, &ch92, &ch93, &ch94, &ch95, 
   &ch96, &ch97, &ch98, &ch99, &ch100, &ch101, &ch102, &ch103, 
   &ch104, &ch105, &ch106, &ch107, &ch108, &ch109, &ch110, &ch111, 
   &ch112, &ch113, &ch114, &ch115, &ch116, &ch117, &ch118, &ch119, 
   &ch120, &ch121, &ch122, &ch123, &ch124, &ch125, &ch126, 0, 
   0, 0, 0, 0, 0, 0, 0, 0, 
   0, 0, 0, 0, 0, 0, 0, 0, 
   0, 0, 0, 0, 0, 0, 0, 0, 
   0, 0, 0, 0, 0, 0, 0, 0, 
   &ch160, &ch161, &ch162, &ch163, &ch164, &ch165, &ch166, &ch167, 
   &ch168, &ch169, &ch170, &ch171, &ch172, &ch173, &ch174, &ch175, 
   &ch176, &ch177, &ch178, &ch179, &ch180, &ch181, &ch182, &ch183, 
   &ch184, &ch185, &ch186, &ch187, &ch188, &ch189, &ch190, &ch191, 
   &ch192, &ch193, &ch194, &ch195, &ch196, &ch197, &ch198, &ch199, 
   &ch200, &ch201, &ch202, &ch203, &ch204, &ch205, &ch206, &ch207, 
   &ch208, &ch209, &ch210, &ch211, &ch212, &ch213, &ch214, &ch215, 
   &ch216, &ch217, &ch218, &ch219, &ch220, &ch221, &ch222, &ch223, 
   &ch224, &ch225, &ch226, &ch227, &ch228, &ch229, &ch230, &ch231, 
   &ch232, &ch233, &ch234, &ch235, &ch236, &ch237, &ch238, &ch239, 
   &ch240, &ch241, &ch242, &ch243, &ch244, &ch245, &ch246, &ch247, 
   &ch248, &ch249, &ch250, &ch251, &ch252, &ch253, &ch254, &ch255
};

const GLUTBitmapFont glutBitmapHelvetica10 = {
   "-Adobe-Helvetica-Medium-R-Normal--10-100-75-75-P-56-ISO8859-1",
   13, 256, chars
};
