
;-----------------------------------------------------
; def Module definition file for the DLL
;-----------------------------------------------------

LIBRARY libGlut INITINSTANCE TERMINSTANCE
PROTMODE
DATA MULTIPLE NONSHARED READWRITE LOADONCALL
CODE LOADONCALL
EXPORTS          ; Names of exported functions and data

;*********************************
;MesaGLUT
        glutCommandFunc		@3001
 	glutCreateWindow	@3002
	glutCreateSubWindow	@3003
	glutDestroyWindow	@3004
	glutDisplayFunc		@3005
	glutEntryFunc		@3006
	glutExtensionSupported	@3007
	glutIdleFunc		@3008
        glutInit		@3009
	glutInitDisplayMode	@3010
	glutInitWindowPosition	@3011
	glutInitWindowSize	@3012
	glutMainLoop		@3013
	glutGet			@3014
	glutGetColor		@3015
	glutGetWindow		@3016
	glutKeyboardFunc	@3017
	glutMouseFunc		@3018
	glutSpecialFunc		@3019
	glutStrokeCharacter	@3020
	glutSetColor		@3021
	glutSetIconTitle	@3022
	glutSetWindow		@3023
	glutSetWindowTitle	@3024

	glutReshapeFunc		@3025
	glutReshapeWindow	@3026


	glutSwapBuffers		@3027
	glutPostRedisplay	@3028
	glutPositionWindow	@3029
	glutVisibilityFunc	@3030
	glutTimerFunc		@3031
	glutMotionFunc		@3032

	glutCreateMenu		@3033
	glutAttachMenu		@3034
	glutDestroyMenu		@3035
	glutAddMenuEntry	@3036
	glutPassiveMotionFunc	@3037

        glutSolidCone		@3038
	glutSolidCube		@3039
	glutSolidDodecahedron	@3040
	glutSolidOctahedron	@3041
	glutSolidIcosahedron	@3042
	glutSolidSphere		@3043
	glutSolidTeapot		@3044
	glutSolidTetrahedron	@3045
	glutSolidTorus		@3046

	glutWireCone		@3047
        glutWireCube		@3048
        glutWireDodecahedron	@3049
	glutWireIcosahedron	@3050
	glutWireOctahedron	@3051
	glutWireSphere		@3052
        glutWireTetrahedron	@3053
        glutWireTorus		@3054
	glutWireTeapot		@3055

;GL_GLEXT_PROTOTYPES	
	glutBitmapCharacter	@3101
        glutBitmap9By15		@3102
        glutBitmapHelvetica10	@3103
        glutBitmapHelvetica12	@3104
        glutBitmapHelvetica18	@3105
        glutBitmapTimesRoman24	@3106
	glutStrokeRoman		@3107
	glutBitmap8By13		@3108
	
;Global constants
;         hab			@12001
;      /* PM anchor block handle         */
;         hpsCurrent		@12002
         XHDC			@12004
         XHWND			@12005



